"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import networkx as nx
import numpy as np
from numba import njit
import psutil

def optimize_allocations(qc):
    from qrisp.permeability import PermeabilityGraph, TerminatorNode

    # G = dag_from_qc(qc, remove_init_nodes=True)
    G = PermeabilityGraph(qc, remove_artificials = True)
    qc_new = qc.clearcopy()

    dealloc_identifier = lambda x: x.op.name == "qb_dealloc"
    alloc_identifer = lambda x: x.op.name == "qb_alloc"

    # mcx_identifier = lambda x : isinstance(x.op, PTControlledOperation) and x.op.base_operation.name == "x"
    # nmcx_identifier = lambda x : not mcx_identifier(x)
    # sub_sort = lambda G : topological_sort(G, prefer = mcx_identifier, delay = nmcx_identifier)
    # for n in topological_sort(G, prefer = dealloc_identifier, delay = alloc_identifer, sub_sort = sub_sort):

    # The sub_sort function takes the indices of the gates in the source qc as
    # a sorting index. This induces a valid topological ordering because of [proof]
    def sub_sort(dag):
        nodes = list(dag.nodes())
        def sort_key(x):
            if isinstance(x, TerminatorNode):
                return 0
            else:
                return x.qc_index
        nodes.sort(key = sort_key)
        return nodes

    for n in topological_sort(G, prefer=dealloc_identifier, delay=alloc_identifer, sub_sort = sub_sort):
        if n.instr:
            qc_new.append(n.instr)

    #The above algorithm does not move allocation gates to their latest possible
    #position (only compared to other deallocation gates)    
    #We therefore manually move the allocation gates to the position right
    #before the first actual instruction on that qubit.
    new_data = []
    delayed_qubit_alloc_dic = {}
    
    for i in range(len(qc_new.data)):
        instr = qc_new.data[i]
        
        if instr.op.name == "qb_alloc":
            delayed_qubit_alloc_dic[instr.qubits[0]] = instr
        else:
            # We sort the qubits in order to prevent non-deterministic compilation behavior
            alloc_qubits = list(set(delayed_qubit_alloc_dic.keys()).intersection(instr.qubits))
            alloc_qubits.sort(key = lambda x : hash(x))
            for qb in alloc_qubits:
                new_data.append(delayed_qubit_alloc_dic[qb])
                del delayed_qubit_alloc_dic[qb]
            
            new_data.append(instr)
    
    for instr in delayed_qubit_alloc_dic.values():
        new_data.append(instr)
    
    qc_new.data = new_data

    return qc_new


# This function performs a topological sort of the graph G where we try to execute any
# deallocation gates as early as possible while still adhering to the topological order.
# We try to perform a depth-first search as described here:
# https://en.wikipedia.org/wiki/Topological_sorting
# According to the Wikipedia page, we are allowed to pick any node as a "starting point"
# of the DF-search, which allows us to modify the algorithm such that it optimizes the
# (de)allocation order. The general idea is to pick the deallocation nodes as
# starting points, where we order them, such that those deallocation nodes that
# "require" the least allocation nodes are executed first. "Require" here means that
# there is a causal relationship between the allocation and deallocation nodes,
# i.e. there is a path from the allocation node to the deallocation node.


# We can thus determine the amount of allocation nodes required for a deallocation node
# n by counting, the amount of allocation nodes in the "ancestors" subgraph of n.
def topological_sort(G, prefer=None, delay=None, sub_sort=nx.topological_sort):
    """
    Function to perform a topological sort on an Unqomp DAG which allows preferring/
    delaying specific types of nodes

    Parameters
    ----------
    G : nx.DiGraph
        The Unqomp DAG.
    prefer : function, optional
        Function which returns True, when presented with an Instruction, that should be
        preferred. The default is the function that returns False on all Operations
    delay : function, optional
        Function which returns True, when presented with an Instruction, that should be
        delayed. The default is the function that returns False on all Operations
    sub_sort : function, optional
        A function which performs a topological sort, which can sorting preferences of
        secondary importance. The default is nx.topological_sort.

    Returns
    -------
    lin : list[UnqompNode]
        The linearized list of UnqompNodes. The init nodes are not included.

    """

    if prefer is None:
        prefer = lambda x: False

    if delay is None:
        delay = lambda x: True

    G = G.copy()

    
    delay_nodes = []
    prefered_nodes = []

    node_list = list(G.nodes())
    
    for i in range(len(node_list)):
        n = node_list[i]
        if n.instr is None:
            continue
        elif prefer(n.instr):
            prefered_nodes.append(i)
        elif delay(n.instr):
            delay_nodes.append(i)
    
    sprs_mat = nx.to_scipy_sparse_array(G, format="csr")
    
    res = toposort_helper(
                    sprs_mat.indptr,
                    sprs_mat.indices.astype(np.int32),
                    len(G),
                    np.array(delay_nodes, dtype = np.int32), 
                    np.array(prefered_nodes, dtype = np.int32))
    
    return [node_list[i] for i in res]


def toposort_helper(indptr, indices, node_amount, delay_nodes, prefered_nodes):
    if psutil.virtual_memory().available/2 < node_amount**2:
        return toposort_helper_sparse(indptr, indices, node_amount, delay_nodes, prefered_nodes)
    else:
        return toposort_helper_dense(indptr, indices, node_amount, delay_nodes, prefered_nodes)


@njit(cache = True)
def toposort_helper_dense(indptr, indices, node_amount, delay_nodes, prefered_nodes):
    # This array returns a graph that reflects all ancestor relations
    # i.e. ancestor_graph[42] is True at all ancestors of node 42
    ancestor_graph = compute_all_ancestors_dense(indptr, indices, node_amount)
    
    n = prefered_nodes.size
    m = delay_nodes.size
    
    # This array will contain the ancestor relations between the
    # prefered/delay nodes
    dependency_matrix = np.zeros((n, m), dtype = np.int8)

    # Fill with information from ancestor_graph    
    for i in range(n):
        for j in range(m):
            if ancestor_graph[prefered_nodes[i], delay_nodes[j]]:
                dependency_matrix[i, j] = 1

    # This array will contain the result
    res = np.zeros(node_amount, dtype = np.int32)
    
    # This array array tracks which nodes have not yet been processed.
    # It is initialized to all True because no nodes have been processed yet.
    remaining_nodes = np.ones(node_amount, dtype = np.int8)
    
    # This integer will contain the amount of nodes that have been processed
    node_counter = 0
    
    if m != 0:
        for i in range(n):
            # For each prefer nodes we compute how many delay nodes are required.
            required_delay_nodes = np.sum(dependency_matrix, axis = 1)
            
            # We determine the prefer node that requires the least delay nodes
            min_node_index = np.argmin(required_delay_nodes)
            prefer_node = prefered_nodes[min_node_index]
            
            # We determine the ancestor nodes of this node that have 
            # not been processed yet
            to_be_processed = ancestor_graph[prefer_node,:] & remaining_nodes
            
            ancestor_indices = np.nonzero(to_be_processed)[0]
            
            # We insert the nodes in the result array.
            # We can assume that order of the nodes induces by their numbering
            # is already a topological ordering. Therefore inserting them in
            # order is also a topological sub sort.
            res[node_counter:node_counter+len(ancestor_indices)] = ancestor_indices
            node_counter += len(ancestor_indices)
            
            # Mark the nodes as processed
            remaining_nodes[ancestor_indices] = 0
            
            
            # Update the depedency matrix: All delay nodes that have been processed
            # don't need to be considered again for all following iterations,
            # we therefore remove them from the other columns
            dependency_matrix = np.clip(dependency_matrix - dependency_matrix[min_node_index, :], 0, 1)
            
            # Finaly we set all nodes in the processed column to 1 so this column
            # is not processed again.
            dependency_matrix[min_node_index, :] = 1

    # Insert the remaining nodes
    res[node_counter:] = np.nonzero(remaining_nodes)[0]
    
    # return the result
    return res


@njit(cache=True)
def compute_all_ancestors_dense(indptr, indices, node_amount):
    # Initialize ancestor sets for all nodes
    ancestors = np.zeros((node_amount, node_amount), dtype=np.bool_)
    
    # Topological sort
    in_degree = np.zeros(node_amount, dtype=np.int64)
    for i in range(node_amount):
        for j in range(indptr[i], indptr[i+1]):
            in_degree[indices[j]] += 1
    
    queue = [i for i in range(node_amount) if in_degree[i] == 0]
    
    while queue:
        node = queue.pop(0)
        ancestors[node, node] = True  # A node is its own ancestor
        
        for i in range(indptr[node], indptr[node+1]):
            child = indices[i]
            # Add current node and its ancestors to child's ancestors
            ancestors[child, :] |= ancestors[node, :]
            in_degree[child] -= 1
            if in_degree[child] == 0:
                queue.append(child)
    
    return ancestors


@njit(cache=True)
def compute_all_ancestors_sparse(indptr, indices, node_amount):
    # Initialize ancestor sets for all nodes using a list of sets
    ancestors = [set([i]) for i in np.arange(node_amount, dtype = np.int32)]
    
    # Topological sort
    in_degree = np.zeros(node_amount, dtype=np.int64)
    for i in range(node_amount):
        for j in range(indptr[i], indptr[i+1]):
            in_degree[indices[j]] += 1
    
    queue = [i for i in range(node_amount) if in_degree[i] == 0]
    
    while queue:
        node = queue.pop(0)
        
        for i in range(indptr[node], indptr[node+1]):
            child = indices[i]
            # Add current node and its ancestors to child's ancestors
            ancestors[child].update(ancestors[node])
            in_degree[child] -= 1
            if in_degree[child] == 0:
                queue.append(child)
    
    return ancestors

# Example usage:
# indptr = np.array([...])
# indices = np.array([...])
# node_amount = ...
# ancestors_sparse = compute_all_ancestors_sparse(indptr, indices, node_amount)

@njit(cache = True)
def toposort_helper_sparse(indptr, indices, node_amount, delay_nodes, prefered_nodes):
    # This array returns a graph that reflects all ancestor relations
    # i.e. ancestor_graph[42] is True at all ancestors of node 42
    ancestor_graph_sparse = compute_all_ancestors_sparse(indptr, indices, node_amount)
    
    n = prefered_nodes.size
    m = delay_nodes.size
    
    # This array will contain the ancestor relations between the
    # prefered/delay nodes
    dependency_matrix = np.zeros((n, m), dtype = np.int8)
    
    delay_nodes_inv = -np.ones(node_amount, np.int32)
    for i in range(len(delay_nodes)):
        delay_nodes_inv[delay_nodes[i]] = i

    for i in range(n):
        for j in ancestor_graph_sparse[prefered_nodes[i]]:
            if delay_nodes_inv[j] != -1:
                dependency_matrix[i, delay_nodes_inv[j]] = 1
    
    # This array will contain the result
    res = np.zeros(node_amount, dtype = np.int32)
    
    # This array array tracks which nodes have not yet been processed.
    # It is initialized to all True because no nodes have been processed yet.
    remaining_nodes = np.ones(node_amount, dtype = np.int8)
    
    # This integer will contain the amount of nodes that have been processed
    node_counter = 0
    
    if m != 0:
        for i in range(n):
            # For each prefer nodes we compute how many delay nodes are required.
            required_delay_nodes = np.sum(dependency_matrix, axis = 1)
            
            # We determine the prefer node that requires the least delay nodes
            min_node_index = np.argmin(required_delay_nodes)
            prefer_node = prefered_nodes[min_node_index]
            
            # We determine the ancestor nodes of this node that have 
            # not been processed yet
            to_be_processed = np.zeros(remaining_nodes.size, dtype = np.int8)
            for k in ancestor_graph_sparse[prefer_node]:
                to_be_processed[k] = remaining_nodes[k]
            
            ancestor_indices = np.nonzero(to_be_processed)[0]
            
            # We insert the nodes in the result array.
            # We can assume that order of the nodes induces by their numbering
            # is already a topological ordering. Therefore inserting them in
            # order is also a topological sub sort.
            res[node_counter:node_counter+len(ancestor_indices)] = ancestor_indices
            node_counter += len(ancestor_indices)
            
            # Mark the nodes as processed
            remaining_nodes[ancestor_indices] = 0
            
            
            # Update the depedency matrix: All delay nodes that have been processed
            # don't need to be considered again for all following iterations,
            # we therefore remove them from the other columns
            dependency_matrix = np.clip(dependency_matrix - dependency_matrix[min_node_index, :], 0, 1)
            
            # Finaly we set all nodes in the processed column to 1 so this column
            # is not processed again.
            dependency_matrix[min_node_index, :] = 1

    # Insert the remaining nodes
    res[node_counter:] = np.nonzero(remaining_nodes)[0]
    
    # return the result
    return res



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

def test_amplitude_amplification():
    from qrisp import QuantumBool, ry, z, amplitude_amplification
    import numpy as np

    def state_function(qb):
        ry(np.pi/8,qb)

    def oracle_function(qb):   
        z(qb)

    qb = QuantumBool()
    state_function(qb)
    assert np.round(qb.get_measurement()[True],2) == 0.04

    amplitude_amplification([qb], state_function, oracle_function, iter=1)
    assert np.round(qb.get_measurement()[True],2) == 0.31

    amplitude_amplification([qb], state_function, oracle_function, iter=1)
    assert np.round(qb.get_measurement()[True],2) == 0.69

    amplitude_amplification([qb], state_function, oracle_function, iter=1)
    assert np.round(qb.get_measurement()[True],2) == 0.96




"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import gphase, rz, cx, conjugate
from qrisp.operators.qubit.visualization import X_,Y_,Z_

from sympy import Symbol
import numpy as np

PAULI_TABLE = {("I","I"):("I",1),("I","X"):("X",1),("I","Y"):("Y",1),("I","Z"):("Z",1),
               ("I","A"):("A",1),("I","C"):("C",1),("I","P0"):("P0",1),("I","P1"):("P1",1),
               ("X","I"):("X",1),("X","X"):("I",1),("X","Y"):("Z",1j),("X","Z"):("Y",(-0-1j)),
               ("X","A"):("P1",1),("X","C"):("P0",1),("X","P0"):("C",1),("X","P1"):("A",1),
               ("Y","I"):("Y",1),("Y","X"):("Z",(-0-1j)),("Y","Y"):("I",1),("Y","Z"):("X",1j),
               ("Y","A"):("P1",1j),("Y","C"):("P0",(-0-1j)),("Y","P0"):("C",1j),
               ("Y","P1"):("A",(-0-1j)),("Z","I"):("Z",1),("Z","X"):("Y",1j),
               ("Z","Y"):("X",(-0-1j)),("Z","Z"):("I",1),("Z","A"):("A",1),
               ("Z","C"):("C",-1),("Z","P0"):("P0",1),("Z","P1"):("P1",-1),
               ("A","I"):("A",1),("A","X"):("P0",1),("A","Y"):("P0",1j),("A","Z"):("A",-1),
               ("A","A"):("I",0),("A","C"):("P0",1),("A","P0"):("I",0),("A","P1"):("A",1),
               ("C","I"):("C",1),("C","X"):("P1",1),("C","Y"):("P1",(-0-1j)),("C","Z"):("C",1),
               ("C","A"):("P1",1),("C","C"):("I",0),("C","P0"):("C",1),("C","P1"):("I",0),
               ("P0","I"):("P0",1),("P0","X"):("A",1),("P0","Y"):("A",(-0-1j)),
               ("P0","Z"):("P0",1),("P0","A"):("A",1),("P0","C"):("I",0),("P0","P0"):("P0",1),
               ("P0","P1"):("I",0),("P1","I"):("P1",1),("P1","X"):("C",1),("P1","Y"):("C",1j),
               ("P1","Z"):("P1",-1),("P1","A"):("I",0),("P1","C"):("C",1),("P1","P0"):("I",0),
               ("P1","P1"):("P1",1)}

#
# QubitTerm
#

class QubitTerm:
    r"""
    
    """

    def __init__(self, factor_dict={}):
        self.factor_dict = dict(factor_dict)
        
        self.hash_value = hash(tuple(sorted(factor_dict.items(), key = lambda x : x[0])))

    def update(self, update_dict):
        self.factor_dict.update(update_dict)
        self.hash_value = hash(tuple(sorted(self.factor_dict.items())))

    def __hash__(self):
        return self.hash_value

    def __eq__(self, other):
        return self.hash_value == other.hash_value
    
    def copy(self):
        return QubitTerm(self.factor_dict.copy())
    
    def is_identity(self):
        return len(self.factor_dict)==0
    
    def binary_representation(self, n):
        x_vector = np.zeros(n, dtype=int)
        z_vector = np.zeros(n, dtype=int)
        for index in range(n):
            curr_factor = self.factor_dict.get(index,"I")
            if curr_factor=="X":
                x_vector[index] = 1
            elif curr_factor=="Z":
                z_vector[index] = 1
            elif curr_factor=="Y":
                x_vector[index] = 1
                z_vector[index] = 1

        return x_vector, z_vector
    
    def serialize(self):
        # This function serializes the QubitTerm in a way that facilitates the 
        # measurement post-processing. To learn more details about the reasoning
        # behind this Serialisation check QubitOperator.conjugation_circuit
        
        # The idea here is to serialize the operator via 3 integers.
        # These integers specify how the energy of a measurement sample should
        # be computed.
        # They are processed in the function "evaluate_observable".
            
        # 1. The Z-int: The binary representation of this integer has a 1 at
        # every digit, where there is a Pauli term in self.
        
        # 2. The and_int: The binary representation has a 1 at every digit, which
        # should participate in an AND evaluation. If the evaluation of the AND
        # does not return True, the energy of this measurement is 0.

        # 3. The ctrl_int: Thi binary representation has a 1 at every digit,
        # which should be flipped before evaluating the AND value.
        
        z_int = 0
        and_int = 0
        ctrl_int = 0
        last_ladder_factor = None
        factor_dict = self.factor_dict
        for i in factor_dict.keys():
            bit = (1<<i)
            
            # Go through the cases and update the appropriate integers
            if factor_dict[i] in ["X", "Y", "Z"]:
                z_int |= bit
                continue
            elif factor_dict[i] == "C":
                ctrl_int |= bit
                last_ladder_factor = bit
                pass
            elif factor_dict[i] == "A":
                last_ladder_factor = bit
                pass
            elif factor_dict[i] == "P0":
                pass
            elif factor_dict[i] == "P1":
                ctrl_int |= bit
                pass
            else:
                continue
            
            and_int |= bit
        
        # The last ladder factor should not participate in the AND but should
        # instead be treated like a Z Operator (see QubitOperator.get_conjugation_circuit
        # to learn why).
        if last_ladder_factor is not None:
            and_int ^= last_ladder_factor
            z_int ^= last_ladder_factor
        
        # Returns a tuple, which contains the relevant integers and a boolean
        # indicating whether this term contains any ladder operators.
        
        return (z_int, and_int, ctrl_int, int(last_ladder_factor is not None))
    
    def to_pauli(self):
        
        from qrisp.operators import X, Y, Z
        res = 1
        
        for i, factor in self.factor_dict.items():
            if factor == "X":
                res *= X(i)
            elif factor == "Y":
                res *= Y(i)
            elif factor == "Z":
                res *= Z(i)
            elif factor == "A":
                res *= (X(i) + 1j*Y(i))*0.5
            elif factor == "C":
                res *= (X(i) - 1j*Y(i))*0.5
            elif factor == "P0":
                res *= (Z(i) + 1)*0.5
            elif factor == "P1":
                res *= (Z(i) - 1)*(-0.5)
        
        return res
    
    def adjoint(self):
        new_factor_dict = {}
        for i, factor in self.factor_dict.items():
            if factor in ["X", "Y", "Z", "P0", "P1"]:
                new_factor_dict[i] = factor
            elif factor == "A":
                new_factor_dict[i] = "C"
            elif factor == "C":
                new_factor_dict[i] = "A"
            
        return QubitTerm(new_factor_dict)
    
    #
    # Simulation
    #
    def simulate(self, coeff, qv, do_change_of_basis = True):

        from qrisp import h, cx, rz, mcp, conjugate, control, QuantumBool, mcx, x, p, s, QuantumEnvironment, gphase
        from qrisp.operators import QubitOperator
        import numpy as np
        # If required, do change of basis. Change of basis here means, that
        # the quantum argument is conjugated with a function that makes the
        # Operator diagonal. Please refer to the comments in QubitOperator.change_of_basis
        # to learn how this works.
        for factor in self.factor_dict.values():
            if factor not in ["I", "Z", "P0", "P1"]:
                qubit_op = QubitOperator({self : coeff})
                with conjugate(qubit_op.change_of_basis)(qv) as diagonal_op:
                    for diagonal_term, coeff in diagonal_op.terms_dict.items():
                        diagonal_term.simulate(coeff, qv)
                return        
        
        # We group the term into 2 types:
        
        # 1. Z-indices, where a Z-Operator is applied
        # 2. Projector indices, i.e. P0 or P1
        
        # Contains the indices of Pauli Z
        Z_indices = []
        
        # Contains the indices of the projection Operators
        projector_indices = []
        # Contains the booleans, which specify whether P0 or P1
        projector_state = []
        
        factor_dict = self.factor_dict
        for i in factor_dict.keys():
            if factor_dict[i] == "Z":
                Z_indices.append(i)
            elif factor_dict[i] == "P0":
                projector_indices.append(i)
                projector_state.append(False)
            elif factor_dict[i] == "P1":
                projector_indices.append(i)
                projector_state.append(True)
        
        # If no non-trivial indices are found, we perform a global phase
        # and are done.
        if len(Z_indices + projector_indices) == 0:
            gphase(coeff, qv[0])
            return
        
        # If there are only projectors, the circuit is a mcp gate
        elif len(Z_indices) == 0:
            
            # Flip the relevant qubits to ensure the right states are projected.
            flip_qubits = [qv[projector_indices[i]] for i in range(len(projector_indices)) if not projector_state[i]]
            
            if len(flip_qubits) == 0:
                env = QuantumEnvironment()
            else:
                env = conjugate(x)(flip_qubits)
            
            # Perform the mcp            
            with env:
                if len(projector_indices) == 1:
                    p(coeff, qv[projector_indices[0]])
                else:
                    mcp(coeff, [qv[i] for i in projector_indices], method = "balauca")
                
            return
        
        # For your reference, here is the circuit that that implements
        # exp(i*phi*H)
        # For 
        # H = A(0)*C(1)*C(2)*Z(3)*X(4)
        
        #           ┌───┐                                                                  ┌───┐
        #     qv.0: ┤ X ├────────────■───────────────────────────────────■─────────────────┤ X ├
        #           └─┬─┘┌───┐       │                                   │            ┌───┐└─┬─┘
        #     qv.1: ──┼──┤ X ├───────o───────────────────────────────────o────────────┤ X ├──┼──
        #             │  └─┬─┘┌───┐  │  ┌───┐┌───┐┌──────────────┐┌───┐  │  ┌───┐┌───┐└─┬─┘  │
        #     qv.2: ──■────■──┤ H ├──┼──┤ X ├┤ X ├┤ Rz(-1.0*phi) ├┤ X ├──┼──┤ X ├┤ H ├──■────■──
        #                     └───┘  │  └─┬─┘└─┬─┘└──────┬───────┘└─┬─┘  │  └─┬─┘└───┘
        #     qv.3: ─────────────────┼────■────┼─────────┼──────────┼────┼────■─────────────────
        #           ┌───┐            │         │         │          │    │  ┌───┐
        #     qv.4: ┤ H ├────────────┼─────────■─────────┼──────────■────┼──┤ H ├───────────────
        #           └───┘          ┌─┴─┐                 │             ┌─┴─┐└───┘
        # hs_anc.0: ───────────────┤ X ├─────────────────■─────────────┤ X ├────────────────────
        #                          └───┘                               └───┘
        
        
        # To implement the ladder operators, we use the ideas from https://arxiv.org/abs/2310.12256
        
        # In it's essence the circuits described there are a conjugation with an inverse GHZ state
        # preparation and a multi controlled RZ gate.
        
        # Since we performed the change of basis however already at the beginning
        # of this function, we don't have to perform the GHZ preparation anymore.
        # We only need to consider the projectors and the Z-indices

        
        # To realize the behavior of the projector indices, not that the
        # Hamiltonian has the form 
        
        # H = P1*H_red
        #   = |1><1|*H_red
        
        # Simulating this H is therefore the controlled version of H_red:
            
        # exp(itH) = |0><0| ID + |1><1| exp(itH_red)
        
        # We can therefore add the projector indices to the computation of
        # the control qubit. If the projector indices are in the wrong state,
        # no phase will be performed, which is precisely what we want
        
        # Determine the control qubits and the control state
        projector_ctrl_state = ""
        projector_qubits = []
        
        for i in range(len(projector_indices)):
            projector_ctrl_state += str(int(projector_state[i]))
            projector_qubits.append(qv[projector_indices[i]])
        
        flip_control_phase = False
        # If the term has no projectors,
        # we don't need a controlled RZ
        if len(projector_qubits) == 0:
            env = QuantumEnvironment()
            control_qubit_available = False
        elif len(projector_qubits) == 1:
            # If there is only one projector qubit, we can use this as control value
            hs_anc = projector_qubits[0]
            control_qubit_available = True
            if not projector_state[i]:
                flip_control_phase = True
            env = QuantumEnvironment()
        else:
            # In any other case, we allocate an additional quantum bool,
            # perform a mcx to compute the control value.
            # To achieve the multi-controlled RZ behavior, we control the RZ
            # on that quantum bool.

            hs_anc = QuantumBool()
            control_qubit_available = True
            
            # Compute the control value
            env = conjugate(mcx)(projector_qubits, 
                                 hs_anc, 
                                 ctrl_state = projector_ctrl_state, 
                                 method = "gray_pt")
        
        
        # Perform the conjugation
        with env:
            
            # The following qubit will be the target of the RZ gate
            anchor_index = Z_indices[-1]
            
            # Before we execute the RZ, we need to deal with the Z indices.
            
            # By inspection we see that acting on qubit 1 & 2 with Z operators,
            # is equivalent to the term (-1)**(q_1 (+) q_2)
            # It therefore suffices to compute the parity value of all the passive
            # qubits and flip the sign of the coefficient based on the parity.

            # To this end we perform several CX gate on-to the anchor qubit.
            # Since the anchor qubit will receive an RZ gate, each bitflip will
            # induce a sign flip of the phase.

            # We achieve this by executing a conjugation with the following function
            def flip_anchor_qubit(qv, anchor_index, Z_indices):
                for i in Z_indices:
                    cx(qv[i], qv[anchor_index])
            
            # Perform the conjugation
            with conjugate(flip_anchor_qubit)(qv, anchor_index = anchor_index, Z_indices = Z_indices[:-1]):
                
                # Perform the controlled RZ
                if control_qubit_available:
                    # Use Selinger's circuit (page 5)
                    with conjugate(cx)(qv[anchor_index], hs_anc):
                        rz(-coeff, qv[anchor_index])
                        if flip_control_phase:
                            coeff = -coeff
                        rz(coeff, hs_anc)
                else:
                    rz(-coeff*2, qv[anchor_index])
                
    
        if len(projector_indices) >= 2:
            # Delete ancilla
            hs_anc.delete()
                    
    #
    # Printing
    #

    def __str__(self):
        # Convert the sympy expression to a string and return it
        expr = self.to_expr()
        return str(expr)
    
    def __repr__(self):
        return str(self)
    
    def non_trivial_indices(self):
        res = set()
        for index, P in self.factor_dict.items():
            res.add(index)
        return res

    def to_expr(self):
        """
        Returns a SymPy expression representing the QubitTerm.

        Returns
        -------
        expr : sympy.expr
            A SymPy expression representing the QubitTerm.

        """

        def to_spin(P, index):
            if P=="I":
                return 1
            if P=="X":
                return X_(index)
            if P=="Y":
                return Y_(index)
            if P=="Z":
                return Z_(index)
            if P=="A":
                return Symbol("A_" + str(index), commutative = False)
            if P=="C":
                return Symbol("C_" + str(index), commutative = False)
            if P=="P0":
                return Symbol("P^0_" + str(index), commutative = False)
            if P=="P1":
                return Symbol("P^1_" + str(index), commutative = False)
        
        expr = 1
        index_list = sorted(list(self.factor_dict.keys()))
        
        for i in index_list:
            expr = expr*to_spin(self.factor_dict[i],str(i))

        return expr

    #
    # Arithmetic
    #

    def __pow__(self, e):
        if isinstance(e, int) and e>=0:
            if e%2==0:
                return QubitTerm({():1})
            else:
                return self
        else:
            raise TypeError("Unsupported operand type(s) for ** or pow(): "+str(type(self))+" and "+str(type(e)))

    def __mul__(self, other):
        result_factor_dict={}
        result_coeff = 1
        a = self.factor_dict
        b = other.factor_dict

        keys = set(a.keys()) | set(b.keys())
        for key in keys:
            factor, coeff = PAULI_TABLE[a.get(key,"I"),b.get(key,"I")]
            if coeff == 0:
                return QubitTerm({}), 0
            if factor != "I":
                result_factor_dict[key]=factor
                result_coeff *= coeff
        return QubitTerm(result_factor_dict), result_coeff
    
    def subs(self, subs_dict):
        """
        
        Parameters
        ----------
        subs_dict : dict
            A dictionary with indices (int) as keys and numbers (int, float, complex) as values.

        Returns
        -------
        QubitTerm
            The resulting QubitTerm.
        result_coeff : int, float, complex
            The resulting coefficient.
        
        """
        result_factor_dict=self.factor_dict.copy()
        result_coeff = 1

        for key, value in subs_dict.items():
            if key in result_factor_dict:
                del result_factor_dict[key]
                result_coeff *= value

        return QubitTerm(result_factor_dict), result_coeff
    
    #
    # Commutativity
    #
    def commutator(self, other):
        from qrisp.operators.qubit import QubitOperator
        
        term_0, coeff_0 = self*other
        term_1, coeff_1 = other*self
        
        temp =  QubitOperator({term_0 : coeff_0}) - QubitOperator({term_1 : coeff_1})
        temp.apply_threshold(0.5)
        
        return temp

    def commute(self, other):
        """
        Checks if two QubitTerms commute.

        """
        a = self.factor_dict
        b = other.factor_dict

        keys = set()
        keys.update(set(a.keys()))
        keys.update(set(b.keys()))

        sign_flip = 1
        
        for key in keys:
            factor_a = PAULI_TABLE[a.get(key, "I"), b.get(key, "I")]
            factor_b = PAULI_TABLE[b.get(key, "I"), a.get(key, "I")]
            
            if factor_a[1] == 0 and factor_b[1] == 0:
                return True
            if factor_a[0] == factor_b[0]:
                if factor_a[1] == factor_b[1]:
                    continue
                elif factor_a[1] == -factor_b[1]:
                    sign_flip *= -1
                    continue
            else:
                return len(self.commutator(other).terms_dict) == 0
                
        return sign_flip == 1

    def commute_qw(self, other):
        """
        Checks if two QubitTerms commute qubit-wise.

        """
        a = self.factor_dict
        b = other.factor_dict

        keys = set()
        keys.update(set(a.keys()))
        keys.update(set(b.keys()))

        for key in keys:
            if not PAULI_TABLE[a.get(key,"I"), b.get(key,"I")] == PAULI_TABLE[b.get(key,"I"), a.get(key,"I")]:
                return False
        return True
    
    def intersect(self, other):
        """
        Checks if two QubitTerms operate on the same qubit.

        """
        return len(set(self.factor_dict.keys()).intersection(other.factor_dict.keys())) != 0
    
    def ladders_agree(self, other):
        """
        Checks if the ladder operators of two QubitTerms operate on the same set of qubits.

        Parameters
        ----------
        other : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        ladder_indices_self = [factor[0] for factor in self.factor_dict.items() if factor[1] in ["A", "C"]]
        ladder_indices_other = [factor[0] for factor in other.factor_dict.items() if factor[1] in ["A", "C"]]
        return set(ladder_indices_self) == set(ladder_indices_other)
        


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp.alg_primitives import QFT
from qrisp.core.gate_application_functions import p, cz
from qrisp.environments import QuantumEnvironment, conjugate
from qrisp.circuit import Operation, PGate, QuantumCircuit


def fourier_adder(a, b, perform_QFT = True):
    """
    In-place adder function based on `this paper <https://arxiv.org/abs/quant-ph/0410184>`__
    Performs the addition
    
    ::
        
        b += a
    

    Parameters
    ----------
    a : int or QuantumVariable or list[Qubit]
        The value that should be added.
    b : QuantumVariable or list[Qubit]
        The value that should be modified in the in-place addition.
    perform_QFT : bool, optional
        If set to ``False``, no QFT is executed. The default is ``True``.

    Examples
    --------
    
    We add two integers:
        
    >>> from qrisp import QuantumFloat, fourier_adder
    >>> a = QuantumFloat(4)
    >>> b = QuantumFloat(4)
    >>> a[:] = 4
    >>> b[:] = 5
    >>> fourier_adder(a,b)
    >>> print(b)
    {9: 1.0}

    """
    
    if perform_QFT:
        env = conjugate(QFT)(b, exec_swap = False)
    else:
        env = QuantumEnvironment()
    
    with env:
        
                    
        b = list(b)
        b = b[::-1]

        
        if isinstance(a, int):
            for i in range(len(b)):
                p(a*np.pi*2**(1+i-len(b)), b[i])
                
        else:
            
            if len(a) > len(b):
                raise Exception("Tried to add QuantumFloat of higher precision onto QuantumFloat of lower precision")
            
            phase_correction_a = np.zeros(len(a))
            phase_correction_b = np.zeros(len(b))
            for j in range(len(a)):
                for i in range(len(b)):    
                    
                    if 1+j+i-len(b) >= 1:
                        continue
                    if 1+j+i-len(b) == 0:
                        cz(a[j], b[i])
                    else:
                        b[i].qs().append(QuasiRZZ(-np.pi*2**(1+j+i-len(b))/2), [a[j], b[i]])
                        
                        phase_correction_a[j] += np.pi*2**(1+j+i-len(b))/2
                        phase_correction_b[i] += np.pi*2**(1+j+i-len(b))/2
            
            for i in range(len(b)):
                if phase_correction_b[i]%(2*np.pi) != 0:
                    p(phase_correction_b[i], b[i])
            for i in range(len(a)):
                if phase_correction_a[i]%(2*np.pi) != 0:
                    p(phase_correction_a[i], a[i])


class QuasiRZZ(Operation):
    
    def __init__(self, angle):
        qc = QuantumCircuit(2)
        qc.cx(qc.qubits[1], qc.qubits[0])
        qc.p(angle, qc.qubits[0])
        qc.cx(qc.qubits[1], qc.qubits[0])
        
        Operation.__init__(self, "quasi_rzz", num_qubits = 2, definition = qc, params = [angle])
        
        self.permeability = {0 : True, 1 : True}
        self.is_qfree = True        
    
    def inverse(self):
        return QuasiRZZ(-self.params[0])
    
    def control(self, num_ctrl_qubits=1, ctrl_state=-1, method=None):
        
        qc = QuantumCircuit(2 + num_ctrl_qubits)
        qc.cx(qc.qubits[-1], qc.qubits[-2])
        qc.append(PGate(self.params[0]).control(num_ctrl_qubits, ctrl_state=ctrl_state, method=method), qc.qubits[:-1])
        qc.cx(qc.qubits[-1], qc.qubits[-2])
        
        res = Operation.control(self, num_ctrl_qubits, ctrl_state=ctrl_state, method=method)
        
        res.definition = qc
        return res



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.core import QuantumVariable
from qrisp.qtypes import QuantumBool
from qrisp.environments import invert
from qrisp.alg_primitives import mcx, cx

# This function performs the Gidney adder from https://arxiv.org/pdf/1709.06648.pdf
def qq_gidney_adder(a, b, c_in = None, c_out = None):
    
    if len(a) != len(b):
        raise Exception("Tried to call Gidney adder with inputs of unequal length")

    if c_out is not None:
        # Convert to qubit if neccessary
        if isinstance(c_out, QuantumBool):
            c_out = c_out[0]
        
        
        b = list(b) + [c_out]
    
    if len(b) == 1:
        cx(a[0],b[0])
        if c_in is not None:
            if isinstance(c_in, QuantumBool):
                c_in = c_in[0]
            
            
            cx(c_in, b[0])
        return
    
    gidney_anc = QuantumVariable(len(b) - 1, name ="gidney_anc*", qs = b[0].qs())
    
    for i in range(len(b)-1):
        
        if i != 0:
            mcx([a[i], b[i]], gidney_anc[i], method = "gidney")
            cx(gidney_anc[i-1], gidney_anc[i])
        elif c_in is not None:
            cx(c_in, b[i])
            cx(c_in, a[i])
            mcx([a[i], b[i]], gidney_anc[i], method = "gidney")
            cx(c_in, gidney_anc[i])
        else:
            mcx([a[i], b[i]], gidney_anc[i], method = "gidney")
            
        if i != len(b) -2:
            cx(gidney_anc[i], b[i+1])
            cx(gidney_anc[i], a[i+1])
    
    cx(gidney_anc[-1], b[-1])
    
    with invert():
        
        for i in range(len(b)-1):
            
            if i != 0:
                mcx([a[i], b[i]], gidney_anc[i], method = "gidney")
                cx(gidney_anc[i-1], gidney_anc[i])
            elif c_in is not None:
                cx(c_in, a[i])
                mcx([a[i], b[i]], gidney_anc[i], method = "gidney")
                cx(c_in, gidney_anc[i])
            else:
                mcx([a[i], b[i]], gidney_anc[i], method = "gidney")
                
             
            if i != len(b) -2:
                cx(gidney_anc[i], a[i+1])
                
    for i in range(len(a)):
        cx(a[i], b[i])
    
    gidney_anc.delete(verify = False)



"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.operators.qubit.qubit_term import QubitTerm
from qrisp.operators.qubit.qubit_operator import QubitOperator

from functools import cache

# Jordan-Wigner annihilation operaror 
#@cache
def a_jw(j):
    d1={i:'Z' for i in range(j)}
    d1[j]='X'
    d2={i:'Z' for i in range(j)}
    d2[j]='Y'
    return QubitOperator({QubitTerm(d1):0.5,QubitTerm(d2):0.5j})

# Jordan-Wigner creation operator 
#@cache
def c_jw(j):
    d1={i:'Z' for i in range(j)}
    d1[j]='X'
    d2={i:'Z' for i in range(j)}
    d2[j]='Y'
    return QubitOperator({QubitTerm(d1):0.5,QubitTerm(d2):-0.5j})
    
@cache
def jordan_wigner(ladder):
    if ladder[1]:
        return c_jw(ladder[0])
    else:
        return a_jw(ladder[0])
    


"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.operators.fermionic.visualization import a_,c_
from qrisp.operators.qubit import A,C,Z

#
# FermionicTerm
#

import numpy as np

class FermionicTerm:
    r"""
    
    """

    def __init__(self, ladder_list=[]):
        
        self.ladder_list = ladder_list
        
        # Compute the hash value such that 
        # terms receive the same hash as their hermitean conjugate
        # this way the FermionicOperator does not have
        # to track both the term and it's dagger
        index_list = [index for index, is_creator in ladder_list]
        is_creator_hash = 0
        for i in range(len(ladder_list)):
            is_creator_hash += ladder_list[i][1]*2**i
        
        self.hash_value = hash(tuple(index_list + [is_creator_hash]))

    def __hash__(self):
        return self.hash_value

    def __eq__(self, other):
        return self.hash_value == other.hash_value
    
    def copy(self):
        return FermionicTerm(self.ladder_list.copy())
    
    def dagger(self):
        return FermionicTerm([(index, not is_creator) for index, is_creator in self.ladder_list[::-1]])
    #
    # Printing
    #

    def __str__(self):
        # Convert the sympy expression to a string and return it
        expr = self.to_expr()
        return str(expr)
    
    def __repr__(self):
        return str(self)

    def to_expr(self):
        """
        Returns a SymPy expression representing the FermionicTerm.

        Returns
        -------
        expr : sympy.expr
            A SymPy expression representing the FermionicTerm.

        """

        def to_ladder(value, index):
            if value:
                return c_(index)
            else:
                return a_(index)
        
        expr = 1
        for index,value in self.ladder_list[::-1]:
            expr *= to_ladder(value,str(index))

        return expr

    #
    # Arithmetic
    #

    def __mul__(self, other):
        result_ladder_list = other.ladder_list + self.ladder_list
        return FermionicTerm(result_ladder_list)
    
    def order(self):
        """
        Not that important, since relevant Hamiltonians (e.g., electronic structure) consist of ordered terms.
        What is needed for trotterization?

        Fermionic commutation relations:

        {a_i,a_j^dagger} = a_i*a_j^dagger + a_j^dagger*a_i = delta_{ij}
        {a_i^dagger,a_j^dagger} = {a_i,a_j} = 0


        Order ladder terms such that 
            1) Raising operators preceed lowering operators
            2) Operators are ordered in descending order of fermionic modes

        Example: a_5^dagger a_2^dagger a_3 a_1

        """
        pass
        
    def sort(self):
        # Sort ladder operators (ladder operator semantics are order in-dependent)
        sorting_list = [-index for index, is_creator in self.ladder_list]
        perm = np.argsort(sorting_list, kind = "stable")
        ladder_list = [self.ladder_list[i] for i in perm]
        flip_sign = permutation_signature(perm)
        
        for i in range(len(ladder_list)-1):
            if ladder_list[i][0] == ladder_list[i+1][0]:
                if ladder_list[i][1] == ladder_list[i+1][1]:
                    flip_sign = 0
                    break
                if i > 1 and ladder_list[i][0] == ladder_list[i-1][0]:
                    flip_sign = 0
                    break
                    
        return FermionicTerm(ladder_list), flip_sign
    
    def fermionic_swap(self, permutation):
        
        permutation = [permutation.index(i) for i in range(len(permutation))]
        new_ladder_list = [(permutation[i], is_creator) for i, is_creator in self.ladder_list]
        
        return FermionicTerm(new_ladder_list)
    
    def intersect(self, other):
        """
        Checks if two QubitTerms operate on the same qubit.

        """
        return len(set([ladder[0] for ladder in self.ladder_list]).intersection([ladder[0] for ladder in other.ladder_list])) != 0
    
    def to_qubit_term(self, mapping_type = "jordan_wigner"):
        if mapping_type == "jordan_wigner":
            res = 1
            for i in range(len(self.ladder_list)):
                temp = 1
                for j in range(self.ladder_list[i][0]):
                    temp = Z(j)*temp
                
                if self.ladder_list[i][1]:
                    temp = temp*C(self.ladder_list[i][0])
                else:
                    temp = temp*A(self.ladder_list[i][0])
            
                res = temp*res
            
            return res
        else:
            raise Exception(f"Don't know fermionic mapping type {mapping_type}")
            
def permutation_signature(perm):
    
    k = 0
    for i in range(len(perm)):
        for j in range(i):
            k += int(perm[i] < perm[j])
            
    return (-1)**(k%2)
            

            
            
            
        
        
    



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import cx, x, control
from qrisp.alg_primitives.arithmetic.adders import fourier_adder
from qrisp.qtypes.quantum_float import QuantumFloat

def q_int_mult(factor_1, factor_2, inpl_adder = fourier_adder, target_qf = None):
    
    if factor_1.size < factor_2.size:
        factor_1, factor_2 = factor_2, factor_1
    
    
    if factor_1.signed or factor_2.signed:
        raise Exception("Signed ripple multiplication currently not supported")

    n = factor_1.size-1
    if target_qf is None:
        s = QuantumFloat(factor_1.size + factor_2.size + 1, 
                         exponent = factor_1.exponent + factor_2.exponent)
        for i in range(factor_2.size):
            cx(factor_2[i], s[i+1+n])

    else:
        target_qf.extend(1, 0)
        s = target_qf
        inpl_adder(factor_2[:s.size-n-1], s[n+1:])
        
    x(s)
    inpl_adder(factor_2, s)
    
    cx(factor_1[0], s)
    for i in range(factor_1.size):
        
        inpl_adder(factor_2[:len(s)-i], s[i:])
        
        if i != factor_1.size-1:
            pass
            cx(factor_1[i], factor_1[i+1])
            cx(factor_1[i+1], s)
            cx(factor_1[i], factor_1[i+1])

    cx(factor_1[-1], s)
    x(s)
    s.reduce(s[0], verify = False)

    return s

def inpl_q_int_mult(operand, cl_int, inpl_adder = fourier_adder):
    if not cl_int%2:
        raise Exception("In-place multiplication with even integers not supported")
    
    for i in range(operand.size-1):
        with control(operand[operand.size-2-i]):
            inpl_adder(cl_int//2, operand[operand.size-1-i:])


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import matplotlib.pyplot as plt
import dill as pickle

class QAOABenchmark:
    """
    This class is a wrapper for representing and evaluating the data collected in the :meth:`.benchmark <qrisp.qaoa.QAOAProblem.benchmark>` method.
    
    Attributes
    ----------
    layer_depth : list[int]
        The amount of QAOA layers for each run.
    circuit_depth : list[int]
        The depth of the compiled circuit of each run.
    qubit_amount : list[int]
        The amount of qubits of the compiled circuit of each run.
    shots : list[int]
        The amount of shots per backend call of each run.
    iterations : list[int]
        The amount of backend calls of each run.
    counts : list[dict]
        The measurement results of the optimized circuit of each run.
    runtime : list[float]
        The amount of time passed (in seconds) of each run.
    optimal_solution : -
        The optimal solution of the problem.
    cost_function : callable
        The classical cost function of the benchmarked problem.
    
    """
    
    def __init__(self, benchmark_data, optimal_solution, cost_function):
        
        self.layer_depth = benchmark_data["layer_depth"]
        self.circuit_depth = benchmark_data["circuit_depth"]
        self.qubit_amount = benchmark_data["qubit_amount"]
        self.shots = benchmark_data["shots"]
        self.iterations = benchmark_data["iterations"]
        self.counts = benchmark_data["counts"]
        self.runtime = benchmark_data["runtime"]
        
        self.optimal_solution = optimal_solution
        self.cost_function = cost_function
        
    def evaluate(self, cost_metric = "oqv", gain_metric = "approx_ratio"):
        r"""
        Evaluates the data in terms of a cost and a gain metric.
        
        **Cost metric**
        
        The default cost metric is overall quantum volume
        
        .. math::
            
            \text{OQV} = \text{circuit_depth} \times \text{qubits} \times \text{shots} \times \text{iterations}
        
        **Gain metric**
        
        By default, two gain metrics are avialable.
        
        The `approximation ratio <https://en.wikipedia.org/wiki/Approximation_algorithm>`_ 
        is a standard quantity in approximation algorithms and can be selected by
        setting ``gain_metric = "approx_ration"``.
        
        The time to solution metric as used in `this paper <http://arxiv.org/abs/2308.02342>`__
        can be selected with ``gain_metric = "tts"``.
        
        Users can implement their own cost/gain metric by calling ``.evaluate`` with a suited function. 
        For more information check the examples.
        

        Parameters
        ----------
        cost_metric : str or callable, optional
            The method to evaluate the cost of each run. The default is "oqv".
        gain_metric : str or callable, optional
            The method to evaluate the gain of each run. The default is "approx_ratio".

        Returns
        -------
        cost_data : list[float]
            A list containing the cost values of each run.
        gain_data : list[float]
            A list containing the gain of each run.
            
        Examples
        --------
        
        We set up a MaxCut instance and perform some benchmarking.
        
        ::
            
            from qrisp import *
            from networkx import Graph
            G = Graph()

            G.add_edges_from([[0,3],[0,4],[1,3],[1,4],[2,3],[2,4]])

            from qrisp.qaoa import maxcut_problem

            max_cut_instance = maxcut_problem(G)

            benchmark_data = max_cut_instance.benchmark(qarg = QuantumVariable(5),
                                       depth_range = [3,4,5],
                                       shot_range = [5000, 10000],
                                       iter_range = [25, 50],
                                       optimal_solution = "11100",
                                       repetitions = 2
                                       )
        
        We now evaluate the cost using the default metrics.
        
        ::
            
            cost_data, gain_data = benchmark_data.evaluate()
            
            print(cost_data[:10])
            #Yields: [17500000, 17500000, 35000000, 35000000, 35000000, 35000000, 70000000, 70000000, 22500000, 22500000]
            print(gain_data[:10])
            #Yields: [0.8425333333333328, 0.9379999999999996, 0.9256666666666667, 0.8816999999999998, 0.764399999999999, 0.6228000000000001, 0.8136000000000001, 0.9213999999999997, 0.8541333333333333, 0.6424333333333333]
        
        To set up a user specified cost metric we create a customized function
        
        ::
            
            def runtime(run_data):
                return run_data["runtime"]
            
            cost_data, gain_data = benchmark_data.evaluate(cost_metric = runtime)
            
        This function extracts the runtime (in seconds) and uses that as a cost metric. 
        The ``run_data`` dictionary contains the following entries:
            
        * ``layer_depth``: The amount of layers
        * ``circuit_depth``: The depth of the compiled circuit as returned by :meth:`.depth <qrisp.QuantumCircuit.depth>` method.
        * ``qubit_amount``: The amount of qubits of the compiled circuit.
        * ``shots``: The amount of shots that have been performed in this run.
        * ``iterations``: The amount of backend calls, that the optimizer was allowed to do.
        * ``counts``: The measurement results as returned by ``qarg.get_measurement()``.
        * ``runtime``: The time (in seconds) that the ``run`` method of :ref:`QAOAProblem` took.
        * ``optimal_solution``: The optimal solution of the problem

        """
        
        if isinstance(cost_metric, str):
            
            if cost_metric == "oqv":
                cost_metric = overall_quantum_volume
            else:
                raise Exception(f"Cost metric {cost_metric} is unknown")
                
            
        if isinstance(gain_metric, str):
            
            if gain_metric == "approx_ratio":
                gain_metric = lambda x : approximation_ratio(x["counts"], self.optimal_solution, self.cost_function)
            
            elif gain_metric == "tts":
                gain_metric = lambda x : time_to_solution(x["counts"], self.optimal_solution, self.cost_function)
            else:
                raise Exception(f"Gain metric {gain_metric} is unknown")
                
        
        cost_data = []
        gain_data = []
        
        for i in range(len(self.layer_depth)):
            
            run_data = {"layer_depth" : self.layer_depth[i],
                         "circuit_depth" : self.circuit_depth[i],
                         "qubit_amount" : self.qubit_amount[i],
                         "shots" : self.shots[i],
                         "iterations" : self.iterations[i],
                         "counts" : self.counts[i],
                         "runtime" : self.runtime[i],
                         "optimal_solution" : self.optimal_solution
                         }
            
            cost_data.append(cost_metric(run_data))
            gain_data.append(gain_metric(run_data))
        
        return cost_data, gain_data
    
    def visualize(self, cost_metric = "oqv", gain_metric = "approx_ratio"):
        """
        Plots the results of :meth:`.evaluate <qrisp.qaoa.QAOABenchmark.evaluate>`.

        Parameters
        ----------
        cost_metric : str or callable, optional
            The method to evaluate the cost of each run. The default is "oqv".
        gain_metric : str or callable, optional
            The method to evaluate the gain of each run. The default is "approx_ratio".

        Examples
        --------
        
        We create a MaxCut instance and benchmark several parameters
        
        ::
            
            from qrisp import *
            from networkx import Graph
            G = Graph()

            G.add_edges_from([[0,3],[0,4],[1,3],[1,4],[2,3],[2,4]])

            from qrisp.qaoa import maxcut_problem

            max_cut_instance = maxcut_problem(G)

            benchmark_data = max_cut_instance.benchmark(qarg = QuantumVariable(5),
                                       depth_range = [3,4,5],
                                       shot_range = [5000, 10000],
                                       iter_range = [25, 50],
                                       optimal_solution = "11100",
                                       repetitions = 2
                                       )
            
        To visualize the results, we call the corresponding method.
        
        ::
            
            benchmark_data.visualize()
            
        .. image:: benchmark_plot.png
            
            
        """
        
        cost_data, gain_data = self.evaluate(cost_metric, gain_metric)
        
        plt.plot(cost_data, gain_data, "x")
        
        if isinstance(cost_metric, str):
            if cost_metric == "oqv":
                cost_name = "Overall quantum volume"
        else:
            cost_name = cost_metric.__name__
            
        if isinstance(gain_metric, str):
            
            if gain_metric == "approx_ratio":
                gain_name = "Approximation ratio"
            elif gain_metric == "tts":
                gain_name = "Time to solution"
        else:
            gain_name = gain_metric.__name__
            
        plt.xlabel(cost_name)
        plt.ylabel(gain_name)
        plt.grid()
        plt.show()
    
    def rank(self, metric = "approx_ratio", print_res = False, average_repetitions = False):
        """
        Ranks the runs of the benchmark according to a given metric.
        
        The default metric is approximation ratio. Similar to :meth:`.evaluate <qrisp.qaoa.QAOABenchmark.evaluate>`,
        the metric can be user specified.

        Parameters
        ----------
        metric : str or callable, optional
            The metric according to which should be ranked. The default is "approx_ratio".

        Returns
        -------
        list[dict]
            List of dictionaries, where the first element has the highest rank.

        Examples
        --------
        
        We create a MaxCut instance and benchmark several parameters
        
        ::
            
            from qrisp import *
            from networkx import Graph
            G = Graph()

            G.add_edges_from([[0,3],[0,4],[1,3],[1,4],[2,3],[2,4]])

            from qrisp.qaoa import maxcut_problem

            max_cut_instance = maxcut_problem(G)

            benchmark_data = max_cut_instance.benchmark(qarg = QuantumVariable(5),
                                       depth_range = [3,4,5],
                                       shot_range = [5000, 10000],
                                       iter_range = [25, 50],
                                       optimal_solution = "11100",
                                       repetitions = 2
                                       )
            
        To rank the results, we call the according method:
        
        ::
            
            print(benchmark_data.rank()[0])
            #Yields: {'layer_depth': 5, 'circuit_depth': 44, 'qubit_amount': 5, 'shots': 10000, 'iterations': 50, 'counts': {'11100': 0.4909, '00011': 0.4909, '00010': 0.002, '11110': 0.002, '00001': 0.002, '11101': 0.002, '10000': 0.0015, '01000': 0.0015, '00100': 0.0015, '11011': 0.0015, '10111': 0.0015, '01111': 0.0015, '00000': 0.0001, '10010': 0.0001, '01010': 0.0001, '11010': 0.0001, '00110': 0.0001, '10110': 0.0001, '01110': 0.0001, '10001': 0.0001, '01001': 0.0001, '11001': 0.0001, '00101': 0.0001, '10101': 0.0001, '01101': 0.0001, '11111': 0.0001, '11000': 0.0, '10100': 0.0, '01100': 0.0, '10011': 0.0, '01011': 0.0, '00111': 0.0}, 'runtime': 1.4269020557403564, 'optimal_solution': '11100'}
            
        """
        
        if isinstance(metric, str):
    
            if metric == "approx_ratio":
                def approx_ratio(x):
                    return approximation_ratio(x["counts"], self.optimal_solution, self.cost_function)
    
                metric = approx_ratio
    
            elif metric == "time_to_sol":
                metric = lambda x: time_to_solution(x["counts"], self.optimal_solution, self.cost_function)
    
        run_data_list = []
    
        if average_repetitions:
            # Create a dictionary to store aggregated averages
            average_dict = {}
    
        for i in range(len(self.layer_depth)):
    
            run_data = {"layer_depth": self.layer_depth[i],
                        "circuit_depth": self.circuit_depth[i],
                        "qubit_amount": self.qubit_amount[i],
                        "shots": self.shots[i],
                        "iterations": self.iterations[i],
                        "runtime": self.runtime[i],
                        "optimal_solution": self.optimal_solution,
                        "counts" : self.counts[i]
                        }
    
            run_data["metric"] = metric(run_data)
            if average_repetitions:
                # Create a unique key based on the parameters
                key = (run_data['layer_depth'], run_data['shots'], run_data['iterations'])
    
                # Add the result to the corresponding key in the dictionary
                if key not in average_dict:
                    average_dict[key] = {
                        'total_metric': 0,
                        'count': 0
                    }
                average_dict[key]['total_metric'] += metric(run_data)
                average_dict[key]['count'] += 1
    
            run_data_list.append(run_data)
    
        if average_repetitions:
            # Calculate the average for each unique parameter combination
            
            temp = list(run_data_list)
            run_data_list = []
            for run_data in temp:
                key = (run_data['layer_depth'], run_data['shots'], run_data['iterations'])
                
                if not key in average_dict:
                    continue
                
                run_data['metric'] = average_dict[key]['total_metric'] / average_dict[key]['count']
                del run_data["counts"]
                del run_data["runtime"]
                run_data_list.append(run_data)
                
                del average_dict[key]
    
        run_data_list.sort(key=lambda x: x["metric"], reverse=True)
        
        if print_res:
            self.print_rank_table(run_data_list, metric.__name__)
    
        return run_data_list

    
    def print_rank_table(self, run_data_list, metric_name):
        """
        Prints a nicely formatted table of the ranked runs.
    
        Parameters
        ----------
        run_data_list : list[dict]
            List of dictionaries containing run data.
        metric : function
            Function to rank the run data
    
        """
        header = ["Rank", metric_name, "Overall QV", "p", "QC depth", "QB count", "Shots", "Iterations"]
        
        
        # Print the header row
        print("{:<5} {:<12} {:<12} {:<4} {:<10} {:<9} {:<7} {:<10}".format(*header))
        print("============================================================================")
        
        for i, run_data in enumerate(run_data_list):
            
            oqv = sci_notation(overall_quantum_volume(run_data), 4)
            metric_value = sci_notation(run_data["metric"], 3)
            
            
            row = [i, metric_value, oqv, run_data["layer_depth"], run_data["circuit_depth"], run_data["qubit_amount"],
                   run_data["shots"], run_data["iterations"]]
            
            # Print each row
            print("{:<5} {:<12} {:<12} {:<4} {:<10} {:<9} {:<7} {:<10}".format(*row))
        
    def save(self, filename):
        """
        Saves the data to the harddrive for later use.

        Parameters
        ----------
        filename : string
            The filename where to save the data.

        Examples
        --------
        
        We create a MaxCut instance and benchmark several parameters
        
        ::
            
            from qrisp import *
            from networkx import Graph
            G = Graph()

            G.add_edges_from([[0,3],[0,4],[1,3],[1,4],[2,3],[2,4]])

            from qrisp.qaoa import maxcut_problem

            max_cut_instance = maxcut_problem(G)

            benchmark_data = max_cut_instance.benchmark(qarg = QuantumVariable(5),
                                       depth_range = [3,4,5],
                                       shot_range = [5000, 10000],
                                       iter_range = [25, 50],
                                       optimal_solution = "11100",
                                       repetitions = 2
                                       )
            
        To save the results, we call the according method.
        
        ::
            
            benchmark_data.save("example.qaoa")
            

        """
        try:
            with open(filename, 'wb') as file:
                pickle.dump(self, file)
            print(f"Benchmark data saved to {filename}")
        except Exception as e:
            print(f"Error saving benchmark data: {e}")

    @classmethod
    def load(cls, filename):
        """
        Loads benchmark data from the harddrive that has been saved by 
        :meth:`.save <qrisp.qaoa.QAOABenchmark.save>`.

        Parameters
        ----------
        filename : string
            The filename to load from.

        Returns
        -------
        obj : QAOABenchmark
            The loaded data.

        Examples
        --------
        
        We assume that the code from the example in :meth:`.save <qrisp.qaoa.QAOABenchmark.save>`
        has been executed and load the corresponding data:
            
        ::
            
            from qrisp.qaoa import QAOABenchmark
            
            benchmark_data = QAOABenchmark.load("example.qaoa")
            
            
        """
        try:
            with open(filename, 'rb') as file:
                obj = pickle.load(file)
            return obj
        except Exception as e:
            print(f"Error loading benchmark data: {e}")
            return None
        
        
    

# create qScore        

def overall_quantum_volume(run_data):
    return run_data["circuit_depth"]*run_data["qubit_amount"]*run_data["shots"]*run_data["iterations"]

def max_five_metric(metric_dict):
    counts = metric_dict["counts"].copy()
    maxfive = sorted(counts , key=counts.get, reverse=True)[:5]
    fivesol = []
    for name, age in counts.items():  
        if name in maxfive:
            fivesol.append((name, age))
    return fivesol


def time_to_solution(counts, optimal_solution, cost_function):
    """
    Parameters
    ----------
    counts : the result dictionary from the QAOA method, contaning the 
                    bitstrings as keys and the counts divided by the number
                    of shots as values.
    optimal_solution: the optimal solution of the problem
    cost_function : Cost Function used to evaluate the optimization

    Returns
    -------
    time to solution measure from http://arxiv.org/abs/2308.02342.
    
    It corresponds to 1/p_opt, where p_opt is the sum of the squared 
    amplitudes associated to the binary strings encoding the optimal solution.

    """
    obj_function = lambda x : cost_function({x : 1})
    optimal_solution_cost = obj_function(optimal_solution)
    
    return 1/sum([v for k,v in counts.items() if obj_function(k)==optimal_solution_cost])


def approximation_ratio(counts, optimal_solution, cost_function):
    """
    Parameters
    ----------
    counts : the result dictionary from the QAOA method, contaning the 
                    bitstrings as keys and the counts divided by the number
                    of shots as values.
    optimal_solution: the optimal solution of the problem
    cost_function : Cost Function used to evaluate the optimization

    Returns
    -------
    approximation ratio measure, commonly used to evaluate the MaxCut Problem

    """
    optimal_cost = cost_function({optimal_solution: 1})
    if optimal_cost < 0:
        return cost_function(counts)/optimal_cost
    else:
        return optimal_cost/cost_function(counts)

def ilog(n, base):
    """
    Find the integer log of n with respect to the base.

    >>> import math
    >>> for base in range(2, 16 + 1):
    ...     for n in range(1, 1000):
    ...         assert ilog(n, base) == int(math.log(n, base) + 1e-10), '%s %s' % (n, base)
    """
    
    if abs(n) < 1:
        n = 1/n
    
    count = 0
    while n >= base:
        count += 1
        n //= base
    return count

def sci_notation(n, prec=3):
    """
    Represent n in scientific notation, with the specified precision.

    >>> sci_notation(1234 * 10**1000)
    '1.234e+1003'
    >>> sci_notation(10**1000 // 2, prec=1)
    '5.0e+999'
    """
    base = 10
    exponent = ilog(n, base)
    
    if abs(n) < 1:
        exponent = -exponent
    
    mantissa = n / base**exponent
    return '{0:.{1}f}e{2:+d}'.format(mantissa, prec, exponent)



"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import QuantumBool, x, mcx
from qrisp.algorithms.qaoa.mixers import controlled_RX_mixer_gen
import itertools


def create_max_indep_set_mixer(G):
    r"""
    Creates the ``controlled_RX_mixer`` for an instance of the maximal independet set problem for a given graph ``G`` following `Hadfield et al. <https://arxiv.org/abs/1709.03489>`_

    The belonging ``predicate`` function indicates if a set can be swapped into the solution.

    Parameters
    ----------
    G : nx.Graph
        The graph for the problem instance.

    Returns
    -------
    function
        A Python function receiving a :ref:`QuantumVariable` and real parameter $\beta$. 
        This function performs the application of the mixer associated to the graph ``G``.

    """

    neighbors_dict = {node: list(G.adj[node]) for node in G.nodes()}

    def predicate(qv,i):
        qbl = QuantumBool()
        if len(neighbors_dict[i])==0:
            x(qbl)
        else:
            mcx([qv[j] for j in neighbors_dict[i]],qbl,ctrl_state='0'*len(neighbors_dict[i]))
        return qbl

    controlled_RX_mixer=controlled_RX_mixer_gen(predicate)

    return controlled_RX_mixer


def create_max_indep_set_cl_cost_function(G):
    """
    Creates the classical cost function for an instance of the maximal independet set problem for a given graph ``G``.

    Parameters
    ----------
    G : nx.Graph
        The Graph for the problem instance.

    Returns
    -------
    cl_cost_function : function
        The classical cost function for the problem instance, which takes a dictionary of measurement results as input.

    """

    def cl_cost_function(res_dic):
        cost = 0
        for state, prob in res_dic.items():
            temp = True
            indices = [index for index, value in enumerate(state) if value == '1']
            combinations = list(itertools.combinations(indices, 2))
            for combination in combinations:
                if combination in G.edges():
                    temp = False
                    break
            if temp: 
                cost += -len(indices)*prob

        return cost

    return cl_cost_function 


def max_indep_set_init_function(qv):
    r"""
    Prepares the initial state $\ket{0}^{\otimes n}$.
    
    Parameters
    ----------
    qv : :ref:`QuantumVariable`
        The quantum argument.
    
    """
    pass


def max_indep_set_problem(G):
    """
    Creates a QAOA problem instance with appropriate phase separator, mixer, and
    classical cost function.

    Parameters
    ----------
    G : nx.Graph
        The graph for the problem instance.

    Returns
    -------
    :ref:`QAOAProblem`
        A QAOA problem instance for MaxIndepSet for a given graph ``G``.

    """        
    from qrisp.qaoa import QAOAProblem, RZ_mixer

    return QAOAProblem(cost_operator=RZ_mixer,
                        mixer=create_max_indep_set_mixer(G),
                        cl_cost_function=create_max_indep_set_cl_cost_function(G),
                        init_function=max_indep_set_init_function)
    



"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from itertools import combinations
import networkx as nx

def transform_max_set_pack_to_mis(problem):
    """
   Transforms a Maximum Set Packing problem instance into a Maximum Independent Set (MIS) problem instance.

    Parameters
    ----------
    problem : list[set]
        A list of sets specifying the problem.

    Returns
    -------
    G : nx.Graph
        The corresponding graph to be solved by an MIS implementation.


    """
    def non_empty_intersection(problem):
        return [(i, j) for (i, s1), (j, s2) in combinations(enumerate(problem), 2) if s1.intersection(s2)]

    # create constraint graph
    G = nx.Graph()
    G.add_nodes_from(range(len(problem)))
    G.add_edges_from(non_empty_intersection(problem))

    return G 




"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

#
# ONLY USED FOR LATEX PRINTING
#

from sympy import Symbol

#
# Pauli symbols (only used for visualization, i.e., LateX printing with SymPy)
#  

class X_(Symbol):

    __slots__ = ("axis","index")

    def __new__(cls, index):
        obj = Symbol.__new__(cls, "%s_%s" %("X",index), commutative=False, hermitian=True)
        obj.index = index
        return obj

class Y_(Symbol):

    __slots__ = ("axis","index")

    def __new__(cls, index):
        obj = Symbol.__new__(cls, "%s_%s" %("Y",index), commutative=False, hermitian=True)
        obj.index = index
        return obj
       
class Z_(Symbol):

    __slots__ = ("axis","index")

    def __new__(cls, index):
        obj = Symbol.__new__(cls, "%s_%s" %("Z",index), commutative=False, hermitian=True)
        obj.index = index
        return obj
    










"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.core.gate_application_functions import h
from qrisp.alg_primitives.qft import QFT


def QPE(
    args, U, precision=None, target=None, iter_spec=False, ctrl_method=None, kwargs={}
):
    r"""
    Evaluates the `quantum phase estimation algorithm
    <https://en.wikipedia.org/wiki/Quantum_phase_estimation_algorithm>`_.

    The unitary to estimate is expected to be given as Python function, which is called
    on ``args``.

    Parameters
    ----------
    args : list
        A list of arguments (could be QuantumVariables) which represent the state,
        the quantum phase estimation is performed on.
    U : function
        A Python function, which will receive the list ``args`` as arguments in the
        course of this algorithm.
    precision : int, optional
        The precision of the estimation. The default is None.
    target : QuantumFloat, optional
        A target QuantumFloat to perform the estimation into. The default is None.
        If given neither a precision nor a target, an Exception will be raised.
    iter_spec : bool, optional
        If set to ``True``, ``U`` will be called with the additional keyword
        ``iter = i`` where ``i`` is the amount of iterations to perform (instead of
        simply calling ``U`` for ``i`` times). The default is False.
    ctrl_method : string, optional
        Allows to specify which method should be used to generate the
        controlled U circuit. For more information check
        :meth:`.control <qrisp.Operation.control>`. The default is None.
    kwargs : dict, optional
        A dictionary of keyword arguments to pass to ``U``. The default is {}.

    Raises
    ------
    Exception
        Tried to perform quantum phase estimation without precision specification.

    Returns
    -------
    res : QuantumFloat
        The QuantumFloat containing the estimated phase as a fraction of $2 \pi$.

    Examples
    --------

    We define a function that applies two phase gates onto its input and estimate the
    applied phase. ::

        from qrisp import p, QuantumVariable, QPE, multi_measurement

        def U(qv):
            x = 0.5
            y = 0.125

            p(x*2*np.pi, qv[0])
            p(y*2*np.pi, qv[1])

        qv = QuantumVariable(2)

        h(qv)

        res = QPE(qv, U, precision = 3)

    >>> multi_measurement([qv, res])
    {('00', 0.0): 0.25,
     ('10', 0.5): 0.25,
     ('01', 0.125): 0.25,
     ('11', 0.625): 0.25}
    >>> res.qs.depth()
    66

    During the phase estimation, ``U`` is called $2^{\text{precision}}$ times. We can
    reduce that number by abusing that we can bundle repeated calls into a single call
    with a modified phase. ::

        def U(qv, iter = None):
            x = 0.5
            y = 0.125

            p(x*2*np.pi*iter, qv[0])
            p(y*2*np.pi*iter, qv[1])

        qv = QuantumVariable(2)

        h(qv)

        res = QPE(qv, U, precision = 3, iter_spec = True)

    >>> multi_measurement([qv, res])
    {('00', 0.0): 0.25,
     ('10', 0.5): 0.25,
     ('01', 0.125): 0.25,
     ('11', 0.625): 0.25}
    >>> res.qs.depth()
    34

    """

    from qrisp import QuantumFloat, control

    if target is None:
        if precision is None:
            raise Exception(
                "Tried to perform quantum phase estimation without"
                "precision specification"
            )
        qpe_res = QuantumFloat(precision, -precision, signed=False)
    else:
        qpe_res = target

    h(qpe_res)

    for i in range(qpe_res.size):
        if iter_spec:
            with control(qpe_res[i], ctrl_method=ctrl_method):
                U(args, iter=2**i, **kwargs)
        else:
            with control(qpe_res[i], ctrl_method=ctrl_method):
                for j in range(2**i):
                    U(args, **kwargs)

    QFT(qpe_res, inv=True)

    return qpe_res



"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numba as nb
import numpy as np
import networkx as nx

@nb.njit(cache = True)
def rlf_coloring(num_vertices, adjacency_matrix):
    colors = np.full(num_vertices, -1)
    uncolored_vertices = np.arange(num_vertices)
    
    current_color = 0
    while uncolored_vertices.size > 0:
        # Find the vertex with the maximum degree
        degrees = np.sum(adjacency_matrix[uncolored_vertices], axis=1)
        max_degree_index = np.argmax(degrees)
        independent_set = [uncolored_vertices[max_degree_index]]
        
        # Build the maximal independent set
        for vertex in uncolored_vertices:
            if np.all(adjacency_matrix[vertex, np.array(independent_set)] == 0):
                independent_set.append(vertex)
        
        # Assign the current color to all vertices in the independent set
        for vertex in independent_set:
            colors[vertex] = current_color
        
        # Remove colored vertices from the list of uncolored vertices
        uncolored_vertices = np.array([v for v in uncolored_vertices if v not in independent_set])
        
        current_color += 1
    
    return colors

@nb.njit(cache = True)
def dsatur_coloring(num_vertices, adjacency_matrix):
    colors = np.full(num_vertices, -1)
    saturation_degrees = np.zeros(num_vertices, dtype=np.int64)
    uncolored_vertices = np.arange(num_vertices)
    
    # Find the vertex with the maximum degree for the first coloring
    degrees = np.sum(adjacency_matrix, axis=1)
    max_degree_vertex = np.argmax(degrees)
    
    # Color the first vertex
    colors[max_degree_vertex] = 0
    uncolored_vertices = uncolored_vertices[uncolored_vertices != max_degree_vertex]
    
    # Update saturation degrees of neighbors
    for neighbor in range(num_vertices):
        if adjacency_matrix[max_degree_vertex, neighbor] == 1:
            saturation_degrees[neighbor] += 1
    
    while uncolored_vertices.size > 0:
        # Find the vertex with the highest saturation degree
        max_saturation = -1
        max_saturation_vertex = -1
        for vertex in uncolored_vertices:
            if saturation_degrees[vertex] > max_saturation:
                max_saturation = saturation_degrees[vertex]
                max_saturation_vertex = vertex
            elif saturation_degrees[vertex] == max_saturation:
                # If tied, choose the vertex with higher degree
                if degrees[vertex] > degrees[max_saturation_vertex]:
                    max_saturation_vertex = vertex
        
        # Find the lowest available color for this vertex
        used_colors = set()
        for neighbor in range(num_vertices):
            if adjacency_matrix[max_saturation_vertex, neighbor] == 1 and colors[neighbor] != -1:
                used_colors.add(colors[neighbor])
        
        available_color = 0
        while available_color in used_colors:
            available_color += 1
        
        # Assign the color
        colors[max_saturation_vertex] = available_color
        
        # Update saturation degrees of uncolored neighbors
        for neighbor in uncolored_vertices:
            if adjacency_matrix[max_saturation_vertex, neighbor] == 1:
                neighbor_colors = set()
                for v in range(num_vertices):
                    if adjacency_matrix[neighbor, v] == 1 and colors[v] != -1:
                        neighbor_colors.add(colors[v])
                saturation_degrees[neighbor] = len(neighbor_colors)
        
        # Remove the colored vertex from uncolored_vertices
        uncolored_vertices = uncolored_vertices[uncolored_vertices != max_saturation_vertex]
    
    return colors

def find_coloring(G):
    adjacency_matrix = nx.to_numpy_array(G)
    
    coloring_1 = rlf_coloring(len(G), adjacency_matrix)
    coloring_2 = dsatur_coloring(len(G), adjacency_matrix)
    
    if np.max(coloring_1) < np.max(coloring_2):
        return coloring_1
    else:
        return coloring_2



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp.core import p, h, cp, cx, x, s, swap

def QFT_inner(qv, exec_swap=True, qiskit_endian=True, inplace_mult=1, use_gms=False, inpl_adder = None):
    from qrisp.misc import is_inv

    qv = list(qv)
    n = len(qv)

    if qiskit_endian:
        qv = qv[::-1]

    if not use_gms:
        from qrisp.environments.quantum_environments import QuantumEnvironment

        env = QuantumEnvironment

    else:
        from qrisp.environments.GMS_environment import GMSEnvironment

        env = GMSEnvironment

    if not is_inv(inplace_mult, n):
        raise Exception(
            "Tried to perform non-invertible inplace multiplication"
            "during Fourier-Transform"
        )


    if inpl_adder is None:
        accumulated_phases = np.zeros(n)
        for i in range(n):
            if accumulated_phases[i] and not use_gms:
                p(accumulated_phases[i], qv[i])
                accumulated_phases[i] = 0
            
            h(qv[i])
    
            if i == n - 1:
                break
    
            with env():
                for k in range(n - i - 1):
                    # cp(inplace_mult * 2 * np.pi / 2 ** (k + 2), qv[k + i + 1], qv[i])
                    
                    if use_gms:
                        cp(inplace_mult * 2 * np.pi / 2 ** (k + 2), qv[i], qv[k + i + 1])
                    else:
                        phase = inplace_mult * 2 * np.pi / 2 ** (k + 2)
                        
                        # cx(qv[k + i + 1], qv[i])
                        # p(-phase/2, qv[i])
                        # cx(qv[k + i + 1], qv[i])
                        
                        
                        cx(qv[i], qv[k + i + 1])
                        p(-phase/2, qv[k + i + 1])
                        cx(qv[i], qv[k + i + 1])
                        
                        
                        accumulated_phases[i] += phase/2
                        accumulated_phases[k + i + 1] += phase/2
        
        
                    
        for i in range(n):
            if accumulated_phases[i] and not use_gms:
                p(accumulated_phases[i], qv[i])
                accumulated_phases[i] = 0
                
    else:
        
        from qrisp import QuantumFloat, conjugate
        reservoir = QuantumFloat(n+1)
        
        def prepare_reservoir(reservoir):
            n = len(reservoir)
            h(reservoir)
            for i in range(n):
                p(np.pi*2**(i-n+1), reservoir[i])
        
        
        with conjugate(prepare_reservoir)(reservoir):

            for i in range(n):
                
                h(qv[i])
        
                if i == n - 1:
                    break
        
                phase_qubits = []
                for k in range(n - i - 1):
                    cx(qv[i], qv[k + i + 1])
                    phase_qubits.append(qv[k + i + 1])
                
                inpl_adder(phase_qubits[::-1], reservoir[-len(phase_qubits)-2:])
                    
                for k in range(n - i - 1):
                    cx(qv[i], qv[k + i + 1])
                
                x(reservoir)
                inpl_adder(phase_qubits[::-1], reservoir[-len(phase_qubits)-2:])
                x(reservoir)
            
            s(qv)
            inpl_adder(qv, reservoir[-n-1:])
        
        reservoir.delete()

    if exec_swap:
        for i in range(n // 2):
            swap(qv[i], qv[n - i - 1])

    return qv



def QFT(
    qv, inv=False, exec_swap=True, qiskit_endian=True, inplace_mult=1, use_gms=False, inpl_adder=None
):
    """
    Performs the quantum fourier transform on the input.

    Parameters
    ----------
    qv : QuantumVariable
        QuantumVariable to transform (in-place).
    inv : bool, optional
        If set to True, the inverse transform will be applied. The default is False.
    exec_swap : bool, optional
        If set to False, the swaps at the end of the transformation will be skipped.
        The default is True.
    qiskit_endian : bool, optional
        If set to False the order of bits will be reversed. The default is True.
    inplace_mult : int, optional
        Allows multiplying the QuantumVariable with an extra factor during the
        transformation. For more information check `the publication
        <https://ieeexplore.ieee.org/document/9815035>`_. The default is 1.
    use_gms : bool, optional
        If set to True, the QFT will be compiled using only GMS gates as entangling
        gates. The default is False.
    inpl_adder : callable, optional
        Uses an adder and a reservoir state to perform the QFT. Read more about 
        it :ref:`here <adder_based_qft>`. The default is None


    """
    from qrisp import gate_wrap, invert

    name = "QFT"
    if not exec_swap:
        name += " no swap"
    if inplace_mult != 1:
        name += " inpl mult " + str(inplace_mult)
    if inpl_adder is not None:
        name += "_adder_supported"

    if inv:
        with invert():
            gate_wrap(permeability=[], is_qfree=False, name=name)(QFT_inner)(
                qv,
                exec_swap=exec_swap,
                qiskit_endian=qiskit_endian,
                inplace_mult=inplace_mult,
                use_gms=use_gms,
                inpl_adder=inpl_adder
            )
    else:
        gate_wrap(permeability=[], is_qfree=False, name=name)(QFT_inner)(
            qv,
            exec_swap=exec_swap,
            qiskit_endian=qiskit_endian,
            inplace_mult=inplace_mult,
            use_gms=use_gms,
            inpl_adder=inpl_adder
        )

    return qv


"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.algorithms.qaoa.problems.maxIndepSet import create_max_indep_set_cl_cost_function, create_max_indep_set_mixer, max_indep_set_init_function
import networkx as nx
from itertools import combinations


def max_set_packing_problem(sets):
    """
    Creates a QAOA problem instance with appropriate phase separator, mixer, and
    classical cost function.

    Parameters
    ----------
    sets : list[set]
        The sets for the problem instance.

    Returns
    -------
    :ref:`QAOAProblem`
        A QAOA problem instance for MaxSetPacking for given ``sets``.

    """        
    from qrisp.qaoa import QAOAProblem, RZ_mixer

    def non_empty_intersection(sets):
        return [(i, j) for (i, s1), (j, s2) in combinations(enumerate(sets), 2) if s1.intersection(s2)]

    # create constraint graph
    G = nx.Graph()
    G.add_nodes_from(range(len(sets)))
    G.add_edges_from(non_empty_intersection(sets))

    return QAOAProblem(cost_operator=RZ_mixer,
                        mixer=create_max_indep_set_mixer(G),
                        cl_cost_function=create_max_indep_set_cl_cost_function(G),
                        init_function=max_indep_set_init_function)


    



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp.alg_primitives import x, cx
from qrisp.alg_primitives.arithmetic.adders.qcla.quantum_quantum.qq_carry_path import qq_calc_carry
from qrisp.alg_primitives.arithmetic.adders.qcla.quantum_quantum.qq_sum_path import qq_sum_path_direct_uncomputation, qq_sum_path
from qrisp.alg_primitives.arithmetic.adders.gidney import gidney_adder
from qrisp.environments import QuantumEnvironment, invert
from qrisp.core import QuantumVariable
from qrisp.circuit import fast_append
from qrisp.misc.utility import redirect_qfunction

verify_manual_uncomputations = np.zeros(1)


# This function performs the in-place addition 
# b += a
# based on the higher radix qcla
# The overall radix can be specified as an exponential of the form
# R = radix_base**radix_exponent
def qq_qcla(a, b, radix_base = 2, radix_exponent = 1, t_depth_reduction = True):
    
    if len(a) > len(b):
        raise Exception("Tried to add QuantumFloat of higher precision onto QuantumFloat of lower precision")
    
    with fast_append():
    # if True:
        R = radix_base**radix_exponent
        
        # The case that a only has a single qubit is simple.
        if len(b) == 1:
            cx(a[0],b[0])
            return
        elif len(b) <= R:
            qcla_anc = QuantumVariable(len(b) - len(a), name = "qcla_anc*", qs = b[0].qs())
            gidney_adder(list(a) + list(qcla_anc), b)
            qcla_anc.delete(verify = bool(verify_manual_uncomputations[0]))
            return
        
        # Calculate the carry
        # Executing within a QuantumEnvironemnt accelerates the uncomputation algorithm
        # because it doesn't have to consider the operations appended outside of this function
        with QuantumEnvironment():
            c = qq_calc_carry(a, b, radix_base, radix_exponent)
        
        if t_depth_reduction:
            qq_sum_path_direct_uncomputation(a,b,c,R)
        else:
            qq_sum_path(a,b,c,R)
            
            # To uncompute the carry we use Drapers strategy
            # CARRY(A,B) = CARRY(A, NOT(A+B))
            # We therefore bitflip the sum
            x(b)
            
            # Contrary to Draper's adder we don't need to uncompute every carry digit.
            # Because of the above equivalence, the carries agree on every digit, so especially
            # on the digits representing the output of the calc_carry function. We can therefore
            # uncompute using calc_carry (even with higher radix) by inverting calc_carry.
            
            with invert():
                #We use the redirect_qfunction decorator to steer the function onto c
                redirect_qfunction(qq_calc_carry)(a, b, radix_base, radix_exponent, target = c)
            
            # Flip the sum back
            x(b)
        
        # In the case R = 1 we can use automatic uncomputation which seems to perform better
        if R == 1:
            c.uncompute()
            return
        # Delete c
        c.delete(verify = bool(verify_manual_uncomputations[0]))
            
        
        


"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import rz, rzz, x
import numpy as np
import copy
from qrisp.algorithms.qiro.qiroproblems.qiro_utils import find_max
from qrisp.algorithms.qiro.qiro_mixers import qiro_controlled_RX_mixer_gen
from qrisp import QuantumBool, mcx
from qrisp.algorithms.qaoa import controlled_RX_mixer_gen


def create_max_indep_replacement_routine(res, problem_updated):
    """
    Creates a replacement routine for the problem structure, i.e., defines the replacement rules. 
    See the `original paper <https://journals.aps.org/prxquantum/abstract/10.1103/PRXQuantum.5.020327>`_ for a description of the update rules.

    Parameters
    ----------
    res : dict
        Result dictionary of QAOA optimization procedure.
    problem_updated : List
        Updates that happened during the QIRO routine. Consits of the updated problem, a list of Qubits which were found to be positively correlated, i.e. part of the problem solution, 
        and a list Qubits which were found to be negatively correlated, i.e. they contradict solution qubits in accordance with the update rules.  

    Returns
    -------
    new_graph : nx.Graph
        Updated graph for the problem instance.
    solutions : list
        Updated set of solutions to the problem.
    sign : int
        The sign of the correlation.
    exclusions : list
        Updated set of exclusions for the problem.
        
    """
    graph = problem_updated[0]
    solutions = problem_updated[1]
    exclusions =  problem_updated[2]

    # for multi qubit correlations
    orig_edges = [list(item) for item in graph.edges()]

    # for single qubit correlations
    orig_nodes = list(graph.nodes())
    
    max_item, sign = find_max(orig_nodes, orig_edges , res, solutions)
    if max_item == None:
        return graph, solutions, 0 ,exclusions

    # create a copy of the graph
    new_graph = copy.copy(graph)

    # we remove nodes from the graph, as suggested by the replacement rules
    # if the item is an int, it is a single node, else it is an edge
    if isinstance(max_item, int):
        # if sign <0 then its mostly appeared as a "1" in the results--> part of solution set
        if sign < 0:
            # remove all adjacent nodes
            to_remove = list(graph.adj[max_item])
            new_graph.remove_nodes_from(to_remove)
            solutions.append(max_item)
            exclusions += to_remove
        # if sign >0 then its mostly appeared as a "0" in the results
        elif sign > 0:
            # remove the node
            new_graph.remove_node(max_item)
            exclusions.append(max_item)

    else:
        if sign > 0:
            # remove both nodes
            new_graph.remove_nodes_from(max_item)
            exclusions += list(max_item)

        elif sign < 0:
            # remove all nodes connected to both nodes 
            intersect = list(set( list(graph.adj[max_item[0]].keys()) ) & set( list(graph.adj[max_item[0]].keys()) ))
            new_graph.remove_nodes_from(intersect)
            exclusions += intersect 

    return new_graph, solutions, sign, exclusions


def create_max_indep_cost_operator_reduced(problem_updated):
    r"""
    Creates the ``cost_operator`` for the problem instance.
    This operator is adjusted to consider qubits that were found to be a part of the problem solution.

    Parameters
    ----------
    problem_updated : List
        Updates that happened during the QIRO routine. Consits of the updated problem, a list of Qubits which were found to be positively correlated, i.e. part of the problem solution, 
        and a list Qubits which were found to be negatively correlated, i.e. they contradict solution qubits in accordance with the update rules.  
    
    Returns
    -------
    cost_operator : function
        A function receiving a :ref:`QuantumVariable` and a real parameter $\gamma$. This function performs the application of the cost operator.

    """

    problem = problem_updated[0]
    solutions = problem_updated[1]

    def cost_operator(qv, gamma):
        for pair in list(problem.edges()):
            #cx(qv[pair[0]], qv[pair[1]])
            rzz(3*gamma, qv[pair[0]], qv[pair[1]])
            rz(-gamma, qv[pair[0]])
            rz(-gamma, qv[pair[1]])
        for i in problem.nodes():
            if not i in solutions:
                rz(gamma, qv[i])

    return cost_operator



def create_max_indep_controlled_mixer_reduced(problem_updated):
    r"""
    Creates the ``controlled_RX_mixer`` for a QIRO instance of the maximal independet set problem for a given graph ``G`` following `Hadfield et al. <https://arxiv.org/abs/1709.03489>`_

    The belonging ``predicate`` function indicates if a set can be swapped into the solution.

    Parameters
    ----------
    problem_updated : List
        Updates that happened during the QIRO routine. Consits of the updated problem, a list of Qubits which were found to be positively correlated, i.e. part of the problem solution, 
        and a list Qubits which were found to be negatively correlated, i.e. they contradict solution qubits in accordance with the update rules.  

    Returns
    -------
    controlled_RX_mixer : function
        A Python function receiving a :ref:`QuantumVariable` and real parameter $\beta$. 
        This function performs the application of the mixer associated to the graph ``G``.

    """

    problem = problem_updated[0]
    solutions = problem_updated[1]
    exclusions =  problem_updated[2]
    neighbors_dict = {node: list(problem.adj[node]) for node in problem.nodes()}

    def qiro_predicate(qv,i):
        qbl = QuantumBool()
        if len(neighbors_dict[i])==0:
            x(qbl)
        else:
            mcx([qv[j] for j in neighbors_dict[i]],qbl,ctrl_state='0'*len(neighbors_dict[i]))
        return qbl

    controlled_RX_mixer=qiro_controlled_RX_mixer_gen(qiro_predicate, solutions+exclusions)

    return controlled_RX_mixer




def qiro_max_indep_set_init_function(solutions =[], exclusions = []):
    r"""
    To be used for the controlled mixer approach of QIRO MIS. Only flips qubits which we found to be a part of the problem soultion.
    
    Parameters
    ----------
    solutions : List
        List of Qubits which were found to be positively correlated, i.e. part of the problem solution
    exclusions : List
        List Qubits which were found to be negatively correlated, i.e. they contradict solution qubits in accordance with the update rules.  
    
    """
    def init_function(qv):
        for i in range(len(qv)):
            if i in solutions:
                x(qv[i])

    return init_function



"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np
from numba import njit

class QubitOperatorMeasurement:
    
    def __init__(self, qubit_op):
        
        hamiltonian = qubit_op.hermitize()
        hamiltonian = hamiltonian.eliminate_ladder_conjugates()
        
        self.groups = hamiltonian.commuting_qw_groups()
        
        self.n = len(self.groups)
        
        # Collect standard deviation
        self.stds = []
        for group in self.groups:
            self.stds.append(np.sqrt(group.to_pauli().hermitize().get_operator_variance()))
            
        N = sum(self.stds)
        self.shots_list = [N*s for s in self.stds]
        
        self.conjugation_circuits = []
        self.measurement_operators = []
        
        for group in self.groups:
            
            conj_circuit, meas_op = group.get_conjugation_circuit()
            
            self.conjugation_circuits.append(conj_circuit)
            self.measurement_operators.append(meas_op)
            
    
    def get_measurement(self, qc, qubit_list, precision, backend):
        
        from qrisp.misc import get_measurement_from_qc
        results = []
        meas_ops = []
        meas_coeffs = []
        
        for i in range(self.n):
            
            group = self.groups[i]
            conjugation_circuit = self.conjugation_circuits[i]
            shots = int(self.shots_list[i]/precision**2)
            
            curr = qc.copy()
            qubits = [qubit_list[i] for i in range(conjugation_circuit.num_qubits())]
            curr.append(conjugation_circuit.to_gate(), qubits)
            
            res = get_measurement_from_qc(curr.transpile(), list(qubit_list), backend, shots)
            results.append(res)
            
            temp_meas_ops = []
            temp_coeff = []
            for term, coeff in group.terms_dict.items():
                temp_meas_ops.append(term.serialize())
                temp_coeff.append(coeff)
                
            meas_coeffs.append(temp_coeff)
            meas_ops.append(temp_meas_ops)
            
        
        samples = create_padded_array([list(res.keys()) for res in results]).astype(np.int64)
        probs = create_padded_array([list(res.values()) for res in results])
        meas_ops = create_padded_array(meas_ops, use_tuples = True).astype(np.int64)
        meas_coeffs = create_padded_array(meas_coeffs)
        
        return evaluate_expectation_jitted(samples, probs, meas_ops, meas_coeffs)
        


def create_padded_array(list_of_lists, use_tuples = False):
    """
    Create a padded numpy array from a list of lists with varying lengths.
    
    Parameters:
    list_of_lists (list): A list of lists with potentially different lengths.
    
    Returns:
    numpy.ndarray: A 2D numpy array with padded rows.
    """
    # Find the maximum length of any list in the input
    max_length = max(len(lst) for lst in list_of_lists)
    
    # Create a padded list of lists
    if not use_tuples:
        padded_lists = [
            lst + [0] * (max_length - len(lst))
            for lst in list_of_lists
        ]
    else:
        padded_lists = [
            lst + [(0,0,0,0)] * (max_length - len(lst))
            for lst in list_of_lists
        ]
        
    
    # Convert to numpy array
    return np.array(padded_lists)


#
# Evaluate expectation
#


def evaluate_expectation(samples, probs, operators, coefficients):
    """
    Evaluate the expectation.
    
    """
    # print(results)
    # print(operators)
    # print(coefficients)
    # raise

    expectation = 0

    for index1,ops in enumerate(operators):
        for index2,op in enumerate(ops):
            for i in range(len(samples[index1])):
                outcome,probability = samples[index1, i], probs[index1, i]
                expectation += probability*evaluate_observable(op,outcome)*np.real(coefficients[index1][index2])
    
    return expectation


def evaluate_observable(observable: tuple, x: int):
    # This function evaluates how to compute the energy of a measurement sample x.
    # Since we are also considering ladder operators, this energy can either be
    # 0, -1 or 1. For more details check out the comments of QubitOperator.get_conjugation_circuit
    
    # The observable is given as tuple, containing for integers and a boolean.
    # To understand the meaning of these integers check QubitTerm.serialize.
    
    # Unwrap the tuple
    z_int, AND_bits, AND_ctrl_state, contains_ladder = observable

    # Compute whether the sign should be sign flipped based on the Z operators
    sign_flip_int = z_int & x
    sign_flip = 0
    while sign_flip_int:
        sign_flip += sign_flip_int & 1
        sign_flip_int >>= 1
    
    # If there is a ladder operator in the term, we need to half the energy 
    # because we want to measure (|110><110| - |111><111|)/2
    if contains_ladder:
        prefactor = 0.5
    else:
        prefactor = 1
    
    # If there are no and bits, we return the result
    if AND_bits == 0:
        return prefactor*(-1)**sign_flip

    # Otherwise we apply the AND_ctrl_state to flip the appropriate bits.
    corrected_x = (x ^ AND_ctrl_state)
    
    # If all bits are in the 0 state the AND is true.
    if corrected_x & AND_bits == 0:
        return prefactor*(-1)**sign_flip
    else:
        return 0
    

evaluate_observable_jitted = njit(cache = True)(evaluate_observable)

@njit(cache = True)
def evaluate_expectation_jitted(samples, probs, operators, coefficients):
    """
    Evaluate the expectation.
    
    """
    # print(results)
    # print(operators)
    # print(coefficients)
    # raise

    expectation = 0

    for index1,ops in enumerate(operators):
        for index2,op in enumerate(ops):
            for i in range(len(samples[index1])):
                outcome,probability = samples[index1, i], probs[index1, i]
                expectation += probability*evaluate_observable_jitted(op,outcome)*np.real(coefficients[index1][index2])
    
    return expectation




"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import app_sb_phase_polynomial
import sympy as sp
import math


def create_maxsat_cost_polynomials(problem):
    """
    Creates a list of polynomials representing the cost function for each clause, and a list of symbols.
    
    Parameters
    ----------
    problem : tuple(int, list[list[int]])
        The number of variables, and the clauses of the maximum satisfiability problem instance.

    Returns
    -------
    cost_polynomials : list[sympy.Expr]
        A list of the cost functions for each clause as SymPy polynomials.
    symbols : list[sympy.Symbol]
        A list of SymPy symbols.

    """

    clauses = problem[1]
    symbols = [sp.Symbol(f"x{i}") for i in range(1,problem[0]+1)]
    cost_polynomials = []
    for clause in clauses:
        C = 1 - sp.prod((1-symbols[index-1]) if index>0 else symbols[-index-1] for index in clause)
        
        cost_polynomials.append(C)

    return cost_polynomials, symbols


def create_maxsat_cl_cost_function(problem):
    """
    Creates the classical cost function for an instance of the maximum satisfiability problem.

    Parameters
    ----------
    problem : tuple(int, list[list[int]])
        The number of variables, and the clauses of the maximum satisfiability problem instance.

    Returns
    -------
    cl_cost_function : function
        The classical cost function for the problem instance, which takes a dictionary of measurement results as input.

    """

    clauses = problem[1]
    def cl_cost_function(res_dic):
        cost = 0
        for state, prob in res_dic.items():
            for clause in clauses:
                cost += -(1-math.prod((1-int(state[index-1])) if index>0 else int(state[-index-1]) for index in clause))*prob

        return cost

    return cl_cost_function 


def create_maxsat_cost_operator(problem):
    r"""
    Creates the cost operator for an instance of the maximum satisfiability problem.
    For a given cost function 

    .. math::

        C(x) = \sum_{\alpha=1}^m C_{\alpha}(x)

    the cost operator is given by $e^{-i\gamma C}$ where $C=\sum_x C(x)\ket{x}\bra{x}$.

    Parameters
    ----------
    problem : tuple(int, list[list[int]])
        The number of variables, and the clauses of the maximum satisfiability problem instance.

    Returns
    -------
    cost_operator : function
        A function receiving a :ref:`QuantumVariable` and a real parameter $\gamma$.
        This function performs the application of the cost operator.

    """
   
    cost_polynomials, symbols = create_maxsat_cost_polynomials(problem)

    def cost_operator(qv, gamma):
        for P in cost_polynomials:
            app_sb_phase_polynomial([qv], -P, symbols, t=gamma)

    return cost_operator


def maxsat_problem(problem):
    """
    Creates a QAOA problem instance with appropriate phase separator, mixer, and
    classical cost function.

    Parameters
    ----------
    problem : tuple(int, list[list[int]])
        The number of variables, and the clauses of the maximum satisfiability problem instance.

    Returns
    -------
    :ref:`QAOAProblem`
        A QAOA problem instance for MaxSat for given a ``problem``.

    """        
    from qrisp.qaoa import QAOAProblem, RX_mixer

    return QAOAProblem(cost_operator=create_maxsat_cost_operator(problem),
                        mixer=RX_mixer,
                        cl_cost_function=create_maxsat_cl_cost_function(problem))